#ifndef MARKER_DETECTION_H
#define MARKER_DETECTION_H

#include "opencv2/opencv.hpp"
#include <stdio.h>

//Structure definition
struct marker_t
{
	std::vector<cv::Point2f> corners;
	int code;	
};

struct orientation_circle_t
{
	std::vector<cv::Point2f> circle;
	cv::Point2f center;		
};

struct possible_marker_t
{
	int contour_index;
	int code;

	std::vector<cv::Point2f> corners;

	std::vector<orientation_circle_t> circles;
	orientation_circle_t orientation_circle;
	std::vector<std::vector<cv::Point2f>> bit_corners;
};

struct threshold_t
{
	int threshold;
	cv::Mat threshold_image;
	cv::Mat original_image;

	std::vector<std::vector<cv::Point>> contours;
	std::vector<cv::Vec4i> hierarchy;

	std::vector<possible_marker_t> possible_markers;
};


bool find_markers(cv::Mat frame, std::vector<marker_t>& markers);

namespace markers
{
	class detector
	{
	private:
		int threshold;
		
		std::vector<std::vector<cv::Point>> contours;
		std::vector<cv::Vec4i> hierarchy;

		cv::Mat drawing;
		cv::Mat threshold_image;
		cv::Mat original_image;

		void find_contours(void);

		void draw_contours(void);

		void draw_squares(std::vector<cv::Point2f> squareCorners);

		void draw_contour(std::vector<cv::Point> contour);

		void find_squares_in_contours(std::vector<possible_marker_t>& possible_markers);

	public:
		detector(int thresh, cv::Mat thresh_img, cv::Mat org_img);

		threshold_t detect_squares(void);

	};


	class identifier
	{
	private:
		threshold_t THRESH;
		cv::Mat drawing;
		std::vector<int> delete_list;

		std::vector<int> actual_markers;

		void searchForMarkers(cv::Mat extractedMarker, int pm_i);

		double getSimilarity( const cv::Mat A, const cv::Mat B );

	public:
		identifier(threshold_t thresh);

		threshold_t identify_markers(void);

	};

	
	/*
	class identifier
	{
	private:
		threshold_t THRESH;
		cv::Mat drawing; 

		std::vector<int> actual_markers;

		void calculate_bit_corners(std::vector<std::vector<cv::Point2f>>& bit_corners, std::vector<cv::Point2f> marker_corners);

		void find_all_circles(int current_square_index, std::vector<std::vector<cv::Point>> contours, std::vector<cv::Vec4i> hierarchy, std::vector<orientation_circle_t>& possible_circles);

		void find_circle_origos(std::vector<orientation_circle_t>& possible_circles);

		bool check_for_orientation_circle(cv::Point2f& p1, cv::Point2f& p2, cv::Point2f& p3, cv::Point2f& p4, std::vector<orientation_circle_t> circles, orientation_circle_t& orient_circ);

		bool check_orientation_circle_contents(orientation_circle_t circle);

		void draw_contour(std::vector<cv::Point> contour);

	public:
		identifier(threshold_t thresh);

		threshold_t identify_markers(void);

	};
	*/

	
	
	class decoder
	{
	private:
		threshold_t THRESH;
		cv::Mat drawing; 
		cv::Mat threshold_img;

		int decode_bit(std::vector<cv::Point2f> bit_corners, int smallest_bit_size);

	public:
		decoder(threshold_t thresh);

		threshold_t decode_marker(void);

	
	};
	
}


//Extra functions
double angle(cv::Point pt1, cv::Point pt2, cv::Point pt0);

std::vector<marker_t> evaluate_results(std::vector<threshold_t> thresh);

#endif
