#include<iostream>
#include<stdio.h>
#include<vector>
#include<ctime>

#include"opencv2/opencv.hpp"

#include"../headers/marker_detection.hpp"
#include"../headers/pose_estimation.hpp"
#include"../headers/image_processing.hpp"
#include"../headers/printing.hpp"
#include"../headers/position_heading_calculation.hpp"
#include"../headers/tracking.hpp"

std::string quay = "NTNU_office";

//Switches for error searching
bool print_image_results = false;
bool printVideoResult = false;

bool run_image(std::string quay_name, std::string image_name);

bool run_video(std::string quay_name, std::string video_name);

void show_frame(cv::Mat image);

int main(int argc, char *argv[])
{
	
	std::string current_file;
	bool ret = false;

	// Check that the program gets a valid input file
	if (argc > 1)
	{
		//Loop through all input-files
		for (int input_index = 1 ; input_index < argc ; input_index++)
		{
			current_file = argv[input_index];
			std::cout << "\nLoaded " << current_file << " into the program." << std::endl;
			

			//Load image into main
			if (current_file.substr(current_file.length()-4) == ".jpg")
			{
				ret = run_image(quay, current_file);
				if (ret)
				{
					std::cout << "Program finished successfully on: " << current_file << std::endl;
				}
				else
				{
					std::cout << "Program did not work on: " << current_file << std::endl;
				}
			}
			

			//Load video into main
			else if ((current_file.substr(current_file.length()-4) == ".avi") || (current_file.substr(current_file.length()-4) == ".mp4"))
			{
				ret = run_video(quay, current_file);
				if (ret)
				{
					std::cout << "Program finished successfully on: " << current_file << std::endl;
				}
				else
				{
					std::cout << "Program did not work on: " << current_file << std::endl;
				}
			}
			//Unable to use file
			else
			{
				std::cout << "Unknown filetype on: " << current_file << std::endl;
			}
		}
	}
	else
	{
		std::cout << "No input file given." << std::endl;
	}
	std::cout << std::endl;
	
	return 0;
}

bool run_image(std::string quay_name, std::string image_name)
{
	position_t ship;
	cv::Mat image, transformationScene2Camera;
	transformationScene2Camera = cv::Mat::eye(4, 4, CV_64F);
	std::vector<marker_t> markers;
	std::vector<cv::Point2f> image_pts;
	std::vector<cv::Point3f> real_pts;
	std::vector<int> detected_codes;
	bool ret = false;

	//Load image
	image = cv::imread(image_name);

	//Check if image was loaded successfully
	if (image.empty())
	{
		std::cout << "No image found with filename: " << image_name << std::endl;
		return false;
	}
	else
	{
		//Find markers in the image
		ret = find_markers(image, markers);
		if (ret)
		{
			cv::Mat detected_marker = pf::print_marker_detected(markers,image);
			show_frame(detected_marker);
			
			//Find the real positions of the markers
			ret = get_real_pts(quay_name, markers, image_pts, real_pts, detected_codes);
			std::cout << "Loaded " << real_pts.size() << " real points" << std::endl;
			std::cout << "Loaded " << image_pts.size() << " image points" << std::endl;
			if (ret)
			{
				
				ret = estimate_pose(image_pts, real_pts, transformationScene2Camera);

				if (ret)
				{
					pose_t cameraPose;

					pf::printTransformationMatrix(transformationScene2Camera);

					calculatePose(cameraPose, transformationScene2Camera);

					pf::printPose(cameraPose);
					
				}
				else
				{
					std::cout << "Unable to estimate the pose of the camera in the scene." << std::endl;
				}
				
			}
			else
			{
				std::cout << "Unable to find correspondence between detected markers and real markers." << std::endl;
				return false;
			}
			
		}
		else
		{
			std::cout << "Function for finding markers failed." << std::endl;
			return false;
		}
		
	}

	if (!print_image_results)
	{
		for (int i = 0 ; i < markers.size() ; i++)
		{
			marker_t current_marker = markers[i];
			pf::print_marker(current_marker);
		}
	}

	return true;
}

bool run_video(std::string quay_name, std::string video_name)
{

	std::ofstream resultFile;
	resultFile.open ("Result.txt", std::ofstream::out | std::ofstream::trunc);
	resultFile << "Frame number 	Psi 	x 	z\n";
	
	cv::Mat frame;
	int frameCount = 0;
	cv::Mat prev_frame;
	cv::Mat gray_frame;
	cv::Mat gray_prev_frame;
	bool ret = false;
	bool initialize = true;
	std::vector<marker_t> markers;
	std::vector<int> detected_codes;
	std::vector<cv::Point2f> image_pts;
	std::vector<cv::Point2f> next_image_pts;
	std::vector<cv::Point3f> real_pts;

	cv::Mat transformationScene2Camera = cv::Mat::eye(4, 4, CV_64F);

	cv::VideoWriter outputVideo; 
	cv::Mat outputFrame;



	int state = 0; //State value of algorithm, 0=initialization, 1=Tracking, 2=Tracking and detection.

	// Load video file
	cv::VideoCapture recording(video_name);
	

	// Validate that the video was loaded successfully
	if (!recording.isOpened())
	{
		std::cout << "Unable to load the requested video: " << video_name << std::endl;
		return false;
	}
	else
	{
		recording >> frame;
		outputVideo.open("videos/Results/Test1.avi", CV_FOURCC('M','J','P','G'), 30, frame.size(), true);
		/*
		if (printVideoResult)
		{
			cv::VideoWriter video("Result.avi",CV_FOURCC('M','J','P','G'),30, frame.size());
		}
		if (printVideoResult)
		{
			video.release();
		}
		*/
		//Video loop
		while ((recording.isOpened()) && (!frame.empty()))
		{
			//cv::resize(frame, frame, cv::Size(), 0.5,0.5);
			frameCount++;

			//clock_t begin = clock();
			img_proc::conv_bw(frame,gray_frame);

			if (state == 0)
			{
				//std::cout << "State is 0" << std::endl;
				//Find markers in the frame
				ret = find_markers(frame, markers);

				if (ret)
				{
					//std::cout << "Detected " << markers.size() << " markers in the frame" << std::endl;
					//cv::Mat detected_marker = pf::print_marker_detected(markers,frame);
					//show_frame(detected_marker);
					//detected_marker.release();

					
					//Find real position of the markers
					if (markers.size()>0)
					{

						ret = get_real_pts(quay_name, markers, image_pts, real_pts, detected_codes);
						markers.clear();
						if (ret)
						{
							//Pose estimation

							ret = estimate_pose(image_pts, real_pts, transformationScene2Camera);
							if (ret)
							{
								/*
								for (int i = 0; i < image_pts.size(); i++)
								{
									std::cout << image_pts.at(i) << std::endl;
								}
								*/

								if (!image_pts.empty())
								{
									pose_t cameraPose;

									//pf::printTransformationMatrix(transformationScene2Camera);

									calculatePose(cameraPose, transformationScene2Camera);

									//pf::printPose(cameraPose);
									resultFile << frameCount << "	" << cameraPose.psi*(180/CV_PI) << "	" << cameraPose.x << "	" << cameraPose.z << "\n";

									//std::cout << "Psi: " << cameraPose.psi << ".  x: " << cameraPose.x << ".  z: " << cameraPose.z << "." << std::endl;
									outputFrame = frame.clone();
									for (int i = 1; i < image_pts.size(); i += 4)
									{ 
										cv::rectangle(outputFrame,image_pts.at(i-1),image_pts.at(i+1), cv::Scalar(0,0,255), 4);
									}
									
									cv::rectangle(outputFrame,cv::Point2f(0,0),cv::Point2f(1079,50),cv::Scalar(255,255,255),-1);
									
									cv::putText(outputFrame, "Frame: " + std::to_string(frameCount) + ". Psi: " + std::to_string(cameraPose.psi*(180/CV_PI)) + ". X: " + std::to_string(cameraPose.x) + ". Z: " + std::to_string(cameraPose.z), cv::Point2f(10,45),cv::FONT_HERSHEY_SIMPLEX, 1, cv::Scalar(0,0,0), 2);
									
									outputVideo.write(outputFrame);

									state = 1;
									
								}
							}
							else
							{
								std::cout << "Pose estimation function failed unexpectedly." << std::endl;
								recording.release();
								return false;
							}
						}
						else
						{
							std::cout << "Function for connecting detected markers with real markers failed unexpectedly." << std::endl;
							recording.release();
							return false;
						}
					}
				}
				else
				{
					std::cout << "Function for finding markers failed unexpectedly." << std::endl;
					recording.release();
					return false;
				}
			}
			else if (state == 1)
			{
				//std::cout << "State is 1" << std::endl;
				ret = tracking::track_points(prev_frame, frame, image_pts, next_image_pts, detected_codes, real_pts);
				if (ret)
				{
					image_pts = next_image_pts;
					//Pose estimation

					ret = estimate_pose(image_pts, real_pts, transformationScene2Camera);
					if (ret)
					{
						/*
						for (int i = 0; i < image_pts.size(); i++)
						{
							std::cout << image_pts.at(i) << std::endl;
						}
						*/

						if (!image_pts.empty())
						{
							pose_t cameraPose;

							//pf::printTransformationMatrix(transformationScene2Camera);

							calculatePose(cameraPose, transformationScene2Camera);

							outputFrame = frame.clone();
							for (int i = 1; i < image_pts.size(); i += 4)
							{ 
								cv::rectangle(outputFrame,image_pts.at(i-1),image_pts.at(i+1), cv::Scalar(0,0,255), 4);
							}
							cv::rectangle(outputFrame,cv::Point2f(0,0),cv::Point2f(1079,50),cv::Scalar(255,255,255),-1);
							cv::putText(outputFrame, "Frame: " + std::to_string(frameCount) + ". Psi: " + std::to_string(cameraPose.psi*(180/CV_PI)) + ". X: " + std::to_string(cameraPose.x) + ". Z: " + std::to_string(cameraPose.z), cv::Point2f(10,45),cv::FONT_HERSHEY_SIMPLEX, 1, cv::Scalar(0,0,0), 2);
							outputVideo.write(outputFrame);

							//pf::printPose(cameraPose);
							resultFile << frameCount << "	" << cameraPose.psi*(180/CV_PI) << "	" << cameraPose.x << "	" << cameraPose.z << "\n";

							//std::cout << "Psi: " << cameraPose.psi << ".  x: " << cameraPose.x << ".  z: " << cameraPose.z << "." << std::endl;
							if (frameCount % 30 == 0)
							{
								state = 2;
							}
							
						}
					}
					else
					{
						std::cout << "Pose estimation function failed unexpectedly." << std::endl;
						recording.release();
						return false;
					}
				}
				else
				{
					std::cout << "Function for connecting detected markers with real markers failed unexpectedly." << std::endl;
					recording.release();
					return false;
				}
			}
			else if (state == 2)
			{
				//std::cout << "State is 2" << std::endl;

				//Tracking
				ret = tracking::track_points(prev_frame, frame, image_pts, next_image_pts, detected_codes, real_pts);

				//Detection
				ret = find_markers(frame, markers);

				//Adding results
				std::vector<int> deleteCodes;
				for (int m_i = 0; m_i < markers.size(); m_i++)
				{
					for (int dm_i = 0; dm_i < detected_codes.size(); dm_i++)
					{
						if (markers[m_i].code == detected_codes[dm_i])
						{
							deleteCodes.push_back(dm_i);
						}
					}
				}
				std::reverse(deleteCodes.begin(),deleteCodes.end());
				for (int dc_i = 0; dc_i < deleteCodes.size(); dc_i++)
				{
					detected_codes.erase(detected_codes.begin()+deleteCodes[dc_i]);
					for (int pt = 3; pt >= 0 ; pt--)
					{
						image_pts.erase(image_pts.begin()+((deleteCodes[dc_i]*4)+pt));
						real_pts.erase(real_pts.begin()+((deleteCodes[dc_i]*4)+pt));
					}
				}
				deleteCodes.clear();

				ret = get_real_pts(quay_name, markers, image_pts, real_pts, detected_codes);
				markers.clear();

				if (ret)
				{
					ret = estimate_pose(image_pts, real_pts, transformationScene2Camera);
					if (ret)
					{
						
						if (!image_pts.empty())
						{
							pose_t cameraPose;

							calculatePose(cameraPose, transformationScene2Camera);

							resultFile << frameCount << "	" << cameraPose.psi*(180/CV_PI) << "	" << cameraPose.x << "	" << cameraPose.z << "\n";

							outputFrame = frame.clone();
							for (int i = 1; i < image_pts.size(); i += 4)
							{ 
								cv::rectangle(outputFrame,image_pts.at(i-1),image_pts.at(i+1), cv::Scalar(0,0,255), 4);
							}
							
							cv::rectangle(outputFrame,cv::Point2f(0,0),cv::Point2f(1079,50),cv::Scalar(255,255,255),-1);
							
							cv::putText(outputFrame, "Frame: " + std::to_string(frameCount) + ". Psi: " + std::to_string(cameraPose.psi*(180/CV_PI)) + ". X: " + std::to_string(cameraPose.x) + ". Z: " + std::to_string(cameraPose.z), cv::Point2f(10,45),cv::FONT_HERSHEY_SIMPLEX, 1, cv::Scalar(0,0,0), 2);
							
							outputVideo.write(outputFrame);

							state = 1;
							
						}
					}
					else
					{
						std::cout << "Pose estimation function failed unexpectedly." << std::endl;
						recording.release();
						return false;
					}
				}
				else
				{
					std::cout << "Function for connecting detected markers with real markers failed unexpectedly." << std::endl;
					recording.release();
					return false;
				}

				outputVideo.write(outputFrame);

			}
			else
			{
				std::cout << "State is unvalid, setting state back to initialization" << std::endl;
				state = 0;
			}

		/*





			//Decide if the program should track previous markers, or find new ones. 
			if (initialize)
			{
				//Find markers in the frame
				ret = find_markers(frame, markers);
				if (ret)
				{
					//Find real position of the markers
					if (markers.size()>0)
					{

						ret = get_real_pts(quay_name, markers, image_pts, real_pts, detected_codes);
						markers.clear();
						if (ret)
						{
							//Pose estimation
							ret = true;//Estimate pose
							if (ret)
							{
								if (!image_pts.empty())
								{
									initialize = false;
								}
							}
							else
							{
								std::cout << "Pose estimation function failed unexpectedly." << std::endl;
								recording.release();
								return false;
							}
						}
						else
						{
							std::cout << "Function for connecting detected markers with real markers failed unexpectedly." << std::endl;
							recording.release();
							return false;
						}
					}
				}
				else
				{
					std::cout << "Function for finding markers failed unexpectedly." << std::endl;
					recording.release();
					return false;
				}
			}
			else
			{
				//Track previous markers and find new ones
				if ( frameCount > 15)
				{
					frameCount = 0;
					//Find markers
					ret = find_markers(frame, markers);
					if(ret)
					{
						ret = get_real_pts(quay_name, markers, image_pts, real_pts, detected_codes);
						if (ret)
						{
							//Pose estimation
						}
						else
						{
							std::cout << "Function for connecting image markers and real markers failed unexpectedly." << std::endl;
							recording.release();
							return false;
						}
					}
					else
					{
						std::cout << "Find markers function failed unexpectedly." << std::endl;
							recording.release();
							return false;
					}
				}
				else
				{
					//track markers
					ret = tracking::track_points(gray_prev_frame, gray_frame, image_pts, next_image_pts, detected_codes, real_pts);
					if (ret)
					{
						image_pts = next_image_pts;
					}
					else
					{
						std::cout << "Tracking function failed unexpectedly." << std::endl;
						recording.release();
						return false;
					}
				}
				if (!image_pts.empty())
				{
					//Pose estimation
					ret = true;//Estimate pose
					if (ret)
					{
						initialize = false;
					}
					else
					{
						std::cout << "Pose estimation function failed unexpectedly." << std::endl;
						recording.release();
						return false;
					}
				}
				else
				{
					std::cout << "Resetting initialization" << std::endl;
					initialize = true;
				}

			}
			clock_t end = clock();
			double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
			std::cout << "The time elapsed for the frame is: " << elapsed_secs << std::endl;
			if(show_video_results)
			{
				cv::Mat frame_with_points = frame.clone();
				for (int point = 0; point < image_pts.size() ; point++)
				{
					if (point % 4 == 0)
					{
  						std::string code = std::to_string(detected_codes[point/4]);
  						cv::putText(frame_with_points, code, image_pts[point], cv::FONT_ITALIC, 2, cv::Scalar(0,0,255), 5);
					}
					cv::circle(frame_with_points, image_pts[point], 5, cv::Scalar(0,0,255), -1);
				}
				img_proc::show_frame(frame_with_points);
			}
			*/
			prev_frame = frame.clone();
			gray_prev_frame = gray_frame.clone();
			recording >> frame;
			//break;
		}

		if (printVideoResult)
		{
			//video.release();
		}
		
	}
	
	recording.release();
	resultFile.close();
	outputVideo.release();
	return true;
}


void show_frame(cv::Mat image)
{
	cv::resize(image, image, cv::Size(), 0.3,0.3);
	cv::imshow("Image", image);
	cv::waitKey(0);
	cv::destroyAllWindows();
}