#include<iostream>
#include<stdio.h>
#include<vector>
#include<cmath>
#include<bitset>

#include"opencv2/opencv.hpp"
#include"../headers/marker_detection.hpp"
#include"../headers/image_processing.hpp"

bool show_images = false;
bool show_all_detected = false;
bool draw_cont = false;
bool draw_sq = false;
bool draw_circle = false;

bool find_markers(cv::Mat frame, std::vector<marker_t>& markers)
{
	cv::Mat bw_frame;
	cv::Mat blur_frame;
	cv::Mat thresh_frame;

	img_proc::conv_bw(frame, bw_frame);
	img_proc::blur_image(bw_frame, blur_frame, 5);

	std::vector<threshold_t> threshold_image_types;

	//Detection loop, goes through a list of thresholds and detects squares inside the image.
	for (int t = 120 ; t < 180 ; t += 10)
	{
		//std::cout << "finding markers with threshold " << t << std::endl;
		//Image processing
		img_proc::threshold_image(blur_frame, thresh_frame, t);
		//cv::erode(thresh_frame,thresh_frame,15);
		cv::dilate(thresh_frame,thresh_frame,300);
		//img_proc::show_image("Thresholded", thresh_frame);

		
		//std::cout << "Image threholded" << std::endl;
		//Detect squares in the image.
		markers::detector frame_detector(t, thresh_frame, frame);
		//std::cout << "frame detector done" << std::endl;
		threshold_t thresh = frame_detector.detect_squares();

		//std::cout << "Detector found " << thresh.possible_markers.size() << " possible squares in the image." << std::endl;
		//std::cout << "detect squares is done" << std::endl;
		//Identify what squares can be markers.
		markers::identifier frame_identifier(thresh);
		//std::cout << "Identifier is made" << std::endl;
		thresh = frame_identifier.identify_markers();
		//std::cout << "Identifier is done" << std::endl;

		//std::cout << "Identifier found " << thresh.possible_markers.size() << " possible markers in the image." << std::endl;
		/*
		markers::decoder frame_decoder(thresh);
		thresh = frame_decoder.decode_marker();
		//std::cout << "Decoder is done" << std::endl;
		*/
		threshold_image_types.push_back(thresh);
	}

	std::vector<marker_t> detected_markers;
	detected_markers = evaluate_results(threshold_image_types);
	for (int dm_i = 0; dm_i < detected_markers.size(); ++dm_i)
	{
		markers.push_back(detected_markers[dm_i]);
	}

	return true;
}


//Marker_detector
markers::detector::detector(int thresh, cv::Mat thresh_img, cv::Mat org_img)
{
	threshold = thresh;
	threshold_image = thresh_img.clone();
	original_image = org_img.clone();
	drawing = original_image.clone();
}

threshold_t markers::detector::detect_squares(void)
{
	threshold_t THRESH; 

	if (!original_image.empty()&& !threshold_image.empty())
	{

		std::vector<possible_marker_t> possible_markers;

		find_contours();
		
		if(draw_cont)
		{
			draw_contours();
			img_proc::show_image("Contours",drawing);
			drawing = original_image.clone();
		}

		find_squares_in_contours(possible_markers);

		if (draw_sq)
		{
			for (int pm_i = 0 ; pm_i < possible_markers.size() ; pm_i++)
			{
				draw_squares(possible_markers[pm_i].corners);
				//draw_contour(contours[possible_markers[pm_i].contour_index]);
			}
			img_proc::show_image("Possible markers",drawing);
		}
		
		THRESH.threshold = threshold;
		THRESH.threshold_image = threshold_image;
		THRESH.original_image = original_image;
		THRESH.contours = contours;
		THRESH.hierarchy = hierarchy;
		THRESH.possible_markers = possible_markers;

	}
	
	return THRESH;
}

void markers::detector::find_contours(void)
{
	cv::Mat inverted_image = 255-threshold_image;
	cv::findContours( inverted_image, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_TC89_KCOS , cv::Point(0, 0));
}

void markers::detector::draw_contours(void)
{
	if (contours.size()>0)
	{
		if (drawing.empty())
		{
			drawing = cv::Mat::zeros(threshold_image.size(), CV_8UC3);
		}
		for( int i = 0; i < contours.size(); i++ )
		{
			cv::Scalar color = cv::Scalar(  rand()&255, rand()&255, rand()&255 );
			drawContours(drawing, contours, i, color, 2, 8, hierarchy, 0, cv::Point());
		}
	}
	else
	{
		std::cout << "Error, no contours to draw in the image." << std::endl;
	}
}

void markers::detector::draw_contour(std::vector<cv::Point> contour)
{
	for (int i = 0; i < contour.size() ; i++)
	{
		if (i == contour.size()-1)
		{
			cv::line(drawing, contour[i], contour[0], cv::Scalar(0,0,255), 4);
		}
		else
		{
			cv::line(drawing, contour[i], contour[i+1], cv::Scalar(0,0,255), 4);
		}	
	}
}

void markers::detector::find_squares_in_contours(std::vector<possible_marker_t>& possible_markers)
{
	int maxArea;
	maxArea = .5*original_image.size().width*original_image.size().height;

	std::vector<cv::Point> approx;

	for(int c_i = 0; c_i < contours.size(); c_i++)
	{
		cv::approxPolyDP(contours[c_i], approx, cv::arcLength(contours[c_i], true)*0.05, true);

		if( approx.size() == 4 && cv::isContourConvex(approx) && cv::contourArea(approx)<maxArea && cv::contourArea(approx)>1000)
		{
			//Ensure that the edges are close to the same length. 
			double edge, maxEdge, minEdge;
			edge = std::sqrt(pow(fabs(approx[3].x-approx[0].x),2)+pow(fabs(approx[3].y-approx[0].y),2));
			maxEdge = edge;
			minEdge = edge;
			for (int i = 0; i < 3; i++)
			{
				edge = std::sqrt(pow(fabs(approx[i].x-approx[i+1].x),2)+pow(fabs(approx[i].y-approx[i+1].y),2));
				minEdge = MIN(edge, minEdge);
				maxEdge = MAX(edge, maxEdge);
			}

			if (minEdge > (0.9*maxEdge))
			{
				double maxCosine = 0;
				for(int i = 2; i < 5; i++)
				{
					double cosine = std::fabs(angle(approx[i%4], approx[i-2], approx[i-1]));
					maxCosine = MAX(maxCosine, cosine);
				}
				if( maxCosine < 0.2 )
				{
					possible_marker_t marker;
					marker.contour_index = c_i;
					for (int a_i = 0 ; a_i < approx.size() ; a_i++)
					{
						marker.corners.push_back(approx[a_i]);
					}
					possible_markers.push_back(marker);
				}
			}
		}
	}
}

void markers::detector::draw_squares(std::vector<cv::Point2f> squareCorners)
{
	for (int i = 0; i < 3 ; i++)
	{
		cv::line(drawing, squareCorners[i], squareCorners[i+1], cv::Scalar(0,0,255), 4);
	}
	cv::line(drawing, squareCorners[3], squareCorners[0], cv::Scalar(0,0,255), 4);
}

//New identifier
//Marker identifier
markers::identifier::identifier(threshold_t thresh)
{

	THRESH = thresh;
	drawing = thresh.original_image.clone();

}

threshold_t markers::identifier::identify_markers(void)
{
	cv::Point2f originalPerspective[4];
	cv::Point2f correctPerspective[4];

	// Lambda Matrix
    cv::Mat_<double> lambda( 2, 4, CV_64FC1 );
    //Input and Output Image;
    cv::Mat grayImage, correctedPerspectiveImage;
     
    //Load the image
    img_proc::conv_bw(THRESH.original_image, grayImage);
    // Set the lambda matrix the same type and size as input
    lambda = cv::Mat::zeros( grayImage.rows, grayImage.cols, grayImage.type() );

    // The 4 points where the mapping is to be done , from top-left in clockwise order
	correctPerspective[0] = cv::Point2f( 0,0);
	correctPerspective[1] = cv::Point2f( grayImage.cols-1,0);
	correctPerspective[2] = cv::Point2f( grayImage.cols-1,grayImage.rows-1);
	correctPerspective[3] = cv::Point2f( 0,grayImage.rows-1);

 	for (int i = 0; i < THRESH.possible_markers.size(); i++)
 	{
 		// The 4 points that select quadilateral on the input , from top-left in clockwise order
		// These four pts are the sides of the rect box used as input
		double area = cv::contourArea(THRESH.possible_markers[i].corners, true);

		if (area >= 0)
		{
			for (int j = 0; j < 4; j++)
			{
				originalPerspective[j] = THRESH.possible_markers[i].corners[j];
			}
		}
		else
		{
			cv::Point2f p1 = THRESH.possible_markers[i].corners[0];
			cv::Point2f p2 = THRESH.possible_markers[i].corners[3];
			cv::Point2f p3 = THRESH.possible_markers[i].corners[2];
			cv::Point2f p4 = THRESH.possible_markers[i].corners[1];
			THRESH.possible_markers[i].corners.at(0) = p1;
			THRESH.possible_markers[i].corners.at(1) = p2;
			THRESH.possible_markers[i].corners.at(2) = p3;
			THRESH.possible_markers[i].corners.at(3) = p4;
			originalPerspective[0] = p1;
			originalPerspective[1] = p2;
			originalPerspective[2] = p3;
			originalPerspective[3] = p4;

			
		}

		// Get the Perspective Transform Matrix i.e. lambda 
		lambda = cv::getPerspectiveTransform( originalPerspective, correctPerspective );

		// Apply the Perspective Transform just found to the src image
		cv::warpPerspective(grayImage,correctedPerspectiveImage,lambda,correctedPerspectiveImage.size() );
		
		cv::resize(correctedPerspectiveImage,correctedPerspectiveImage, cv::Size(100,100));

 		//img_proc::show_image("Transformed", correctedPerspectiveImage);

 		searchForMarkers(correctedPerspectiveImage, i);

 	}
 	for (int dl_i = int(delete_list.size())-1 ; dl_i >=0 ; dl_i--)
	{
		THRESH.possible_markers.erase(THRESH.possible_markers.begin()+delete_list[dl_i]);
	}
	return THRESH;
}

void markers::identifier::searchForMarkers(cv::Mat extractedMarker, int pm_i)
{
	bool identified = false;
	for (int marker_i = 0; marker_i < 7; marker_i++)
	{
		std::string marker_binary = std::bitset<3>(marker_i).to_string();
		cv::Mat markerImage = cv::imread("markerLibrary/library/Marker_"+marker_binary+".jpg",cv::IMREAD_GRAYSCALE);
		img_proc::threshold_image(markerImage, markerImage, 150);
		img_proc::threshold_image(extractedMarker, extractedMarker, THRESH.threshold);
		for (int rotation_i = 0; rotation_i < 4; rotation_i++)
		{
			
			cv::Mat output;
			cv::bitwise_xor(markerImage, extractedMarker, output);
			int numb_unequal = cv::countNonZero(output);
			double percentEqual = 100.0-((double)numb_unequal/100.0);

			if (percentEqual >= 90)
			{
				//img_proc::show_image("marker ",markerImage);
				//img_proc::show_image("ex marker ",extractedMarker);
				//std::cout << "Marker " << marker_binary << " detected with " << percentEqual << " accuracy" << std::endl;
				//std::cout << "Rotation is " << rotation_i << std::endl;
				if (rotation_i == 0)
				{
					cv::Point2f p1 = THRESH.possible_markers[pm_i].corners[1];
					cv::Point2f p2 = THRESH.possible_markers[pm_i].corners[0];
					cv::Point2f p3 = THRESH.possible_markers[pm_i].corners[3];
					cv::Point2f p4 = THRESH.possible_markers[pm_i].corners[2];
					THRESH.possible_markers[pm_i].corners.at(0) = p1;
					THRESH.possible_markers[pm_i].corners.at(1) = p2;
					THRESH.possible_markers[pm_i].corners.at(2) = p3;
					THRESH.possible_markers[pm_i].corners.at(3) = p4;
				}
				else if (rotation_i == 1)
				{
					cv::Point2f p1 = THRESH.possible_markers[pm_i].corners[2];
					cv::Point2f p2 = THRESH.possible_markers[pm_i].corners[1];
					cv::Point2f p3 = THRESH.possible_markers[pm_i].corners[0];
					cv::Point2f p4 = THRESH.possible_markers[pm_i].corners[3];
					THRESH.possible_markers[pm_i].corners.at(0) = p1;
					THRESH.possible_markers[pm_i].corners.at(1) = p2;
					THRESH.possible_markers[pm_i].corners.at(2) = p3;
					THRESH.possible_markers[pm_i].corners.at(3) = p4;
				}
				else if (rotation_i == 2)
				{
					cv::Point2f p1 = THRESH.possible_markers[pm_i].corners[3];
					cv::Point2f p2 = THRESH.possible_markers[pm_i].corners[2];
					cv::Point2f p3 = THRESH.possible_markers[pm_i].corners[1];
					cv::Point2f p4 = THRESH.possible_markers[pm_i].corners[0];
					THRESH.possible_markers[pm_i].corners.at(0) = p1;
					THRESH.possible_markers[pm_i].corners.at(1) = p2;
					THRESH.possible_markers[pm_i].corners.at(2) = p3;
					THRESH.possible_markers[pm_i].corners.at(3) = p4;
				}
				else
				{
					cv::Point2f p1 = THRESH.possible_markers[pm_i].corners[0];
					cv::Point2f p2 = THRESH.possible_markers[pm_i].corners[3];
					cv::Point2f p3 = THRESH.possible_markers[pm_i].corners[2];
					cv::Point2f p4 = THRESH.possible_markers[pm_i].corners[1];
					THRESH.possible_markers[pm_i].corners.at(0) = p1;
					THRESH.possible_markers[pm_i].corners.at(1) = p2;
					THRESH.possible_markers[pm_i].corners.at(2) = p3;
					THRESH.possible_markers[pm_i].corners.at(3) = p4;
				}
				if (marker_i == 0) {THRESH.possible_markers[pm_i].code = 0;}
				else if (marker_i == 1) {THRESH.possible_markers[pm_i].code = 1;}
				else if (marker_i == 2) {THRESH.possible_markers[pm_i].code = 10;}
				else if (marker_i == 3) {THRESH.possible_markers[pm_i].code = 11;}
				else if (marker_i == 4) {THRESH.possible_markers[pm_i].code = 100;}
				else if (marker_i == 5) {THRESH.possible_markers[pm_i].code = 101;}
				else if (marker_i == 6) {THRESH.possible_markers[pm_i].code = 110;}
				else if (marker_i == 7) {THRESH.possible_markers[pm_i].code = 111;}

				identified = true;
				break;
			}

			cv::rotate(markerImage, markerImage, cv::ROTATE_90_CLOCKWISE);
		}
		if (identified)
		{
			break;
		}
	}
	if (!identified)
	{
		delete_list.push_back(pm_i);
	}
}


/*
Old part

//Marker identifier
markers::identifier::identifier(threshold_t thresh)
{
	THRESH = thresh;
	drawing = thresh.original_image.clone();
}

threshold_t markers::identifier::identify_markers(void)
{
	std::vector<int> delete_list;

	//std::cout << "Size of possible_markers is " << THRESH.possible_markers.size() << std::endl;

	for (int ps_i = 0 ; ps_i < THRESH.possible_markers.size() ; ps_i++)
	{
		//std::cout << "Current index of ps_i is: " << ps_i << std::endl;
		int square_index = THRESH.possible_markers[ps_i].contour_index;

		std::vector<cv::Point2f> circle_origos;

		int current_child = THRESH.hierarchy[square_index][2];
		bool bottom_reached = false;

		if (current_child != -1)
		{

			find_all_circles(square_index, THRESH.contours, THRESH.hierarchy, THRESH.possible_markers[ps_i].circles);
			
			//If there are no circles the rest can be skipped
			if (THRESH.possible_markers[ps_i].circles.size() != 0)
			{
				std::cout << "Found " << THRESH.possible_markers[ps_i].circles.size() << " circles." << std::endl;
				//Find origo for all detected circles.
				find_circle_origos(THRESH.possible_markers[ps_i].circles);
				//std::cout << "find_circle_origos is done" << std::endl;
				//Check if any circle fits the orientation circle
				cv::Point2f p1, p2, p3, p4;

				p1 = THRESH.possible_markers[ps_i].corners[0];
				p2 = THRESH.possible_markers[ps_i].corners[1];
				p3 = THRESH.possible_markers[ps_i].corners[2];
				p4 = THRESH.possible_markers[ps_i].corners[3];
				//std::cout << "Points p1 to p4 is made" << std::endl;
				bool ret = check_for_orientation_circle(p1, p2, p3, p4, THRESH.possible_markers[ps_i].circles, THRESH.possible_markers[ps_i].orientation_circle);
				//std::cout << "check_for_orientation_circle is done" << std::endl;

				if (ret)
				{
					//std::cout << "Second if" << std::endl;
					ret = check_orientation_circle_contents(THRESH.possible_markers[ps_i].orientation_circle);

					if (ret)
					{
						//std::cout << "Third if" << std::endl;
						THRESH.possible_markers[ps_i].corners.clear();
						std::vector<std::vector<cv::Point2f>> bit_corners;
						THRESH.possible_markers[ps_i].corners.push_back(p1);
						THRESH.possible_markers[ps_i].corners.push_back(p2);
						THRESH.possible_markers[ps_i].corners.push_back(p3);
						THRESH.possible_markers[ps_i].corners.push_back(p4);
						calculate_bit_corners(bit_corners, THRESH.possible_markers[ps_i].corners);
						for (int bit = 0; bit < bit_corners.size(); bit++)
						{
							THRESH.possible_markers[ps_i].bit_corners.push_back(bit_corners[bit]);
						}
					}
					else
					{
						delete_list.push_back(ps_i);
					}
				}
				else
				{
					delete_list.push_back(ps_i);
				}
			}
			else
			{
				delete_list.push_back(ps_i);
			}
		}
		else
		{
			delete_list.push_back(ps_i);
		}
	}
	//std::cout << "End of for loop of possible markers." << std::endl;
	for (int dl_i = int(delete_list.size())-1 ; dl_i >=0 ; dl_i--)
	{
		THRESH.possible_markers.erase(THRESH.possible_markers.begin()+delete_list[dl_i]);
	}
	return THRESH;
}


void markers::identifier::find_all_circles(int current_square_index, std::vector<std::vector<cv::Point>> contours, std::vector<cv::Vec4i> hierarchy, std::vector<orientation_circle_t>& possible_circles)
{
	
	std::vector<int> list_of_contours;
	std::vector<int> new_list_of_contours;

	int current_child = hierarchy[current_square_index][2];
	bool bottom_reached = false;
	//std::cout << "Before first while" << std::endl;
	//Find all contours in the first layer inside the square
	while (current_child != -1)
	{
		new_list_of_contours.push_back(current_child);
		current_child = hierarchy[current_child][0];
	}
	//std::cout << "Before second while" << std::endl;
	//Go through all the contours inside the square and find the circles
	while (!bottom_reached)
	{
		list_of_contours = new_list_of_contours;
		new_list_of_contours.clear();
		//std::cout << "Before first for loop" << std::endl;
		//Go through current list of contours and save the circles
		for (int c_i = 0 ; c_i < list_of_contours.size() ; c_i++)
		{
			std::vector<cv::Point2f> approx;
			cv::approxPolyDP(contours[list_of_contours[c_i]],approx,cv::arcLength(contours[list_of_contours[c_i]], true)*0.02, true);
			if (approx.size() >= 8 && cv::isContourConvex(approx))
			{
				orientation_circle_t or_circle;
				for (int a = 0; a < approx.size(); a++)
				{
					or_circle.circle.push_back(approx[a]);
				}
				possible_circles.push_back(or_circle);

				if (draw_circle)
				{
					draw_contour(contours[list_of_contours[c_i]]);
					img_proc::show_image("Circle contour found", drawing);
				}
			}
		}

		//std::cout << "Before second for loop" << std::endl;
		//Find all the children of the contours in the current list
		for (int c_i = 0 ; c_i < list_of_contours.size() ; c_i++)
		{
			if (hierarchy[list_of_contours[c_i]][2] != -1)
			{
				new_list_of_contours.push_back(hierarchy[list_of_contours[c_i]][2]);
			}
		}

		//std::cout << "Before third for loop" << std::endl;
		//std::cout << "Size of new_list_of_contours is " << new_list_of_contours.size() << std::endl;
		//Find all the next contours of the contours found above
		for (int nc_i = 0 ; nc_i < list_of_contours.size() ; nc_i++)
		{
			//std::cout << "nc_i is now " << nc_i << std::endl;
			current_child = hierarchy[new_list_of_contours[nc_i]][0];
			while (current_child != -1)
			{
				new_list_of_contours.push_back(current_child);
				current_child = hierarchy[current_child][0];
			}
			if(nc_i > 20)
			{
				break;
			}
		}
		//std::cout << "After third for loop" << std::endl;
		//Clear the list of the previous layer of the contours
		list_of_contours.clear();

		//If there are no contours left to check the bottom is reached and there are no more contours to check inside the square.
		if (new_list_of_contours.size() == 0)
		{
			bottom_reached = true;
		}
	}
}


void markers::identifier::find_circle_origos(std::vector<orientation_circle_t>& possible_circles)
{

	//Go through all circles and calculate the origos
	for (int pc_i = 0 ; pc_i < possible_circles.size() ; pc_i++)
	{
		std::vector<cv::Point2f> current_contour = possible_circles[pc_i].circle;
		
		//Sum together all values of the points in the circle
		int x_sum = 0;
		int y_sum = 0;
		for (int c_i = 0 ; c_i < current_contour.size() ; c_i++)
		{
			x_sum += current_contour[c_i].x;
			y_sum += current_contour[c_i].y;
		}

		//The origo is the sum divided by the number of points
		float origo_x = float(float(x_sum)/float(current_contour.size()));
		float origo_y = float(float(y_sum)/float(current_contour.size()));

		//Add the origo to the list of origos
		possible_circles[pc_i].center = cv::Point2f(origo_x,origo_y);
	}
}

bool markers::identifier::check_for_orientation_circle(cv::Point2f& p1, cv::Point2f& p2, cv::Point2f& p3, cv::Point2f& p4, std::vector<orientation_circle_t> circles, orientation_circle_t& orient_circ)
{

	bool draw_circle_origos = false;

	cv::Point2f pos1, pos2, pos3, pos4;
	cv::Point2f m1, m2, m3, m4;
	
	m1.x = p1.x+(7.0/24.0)*(p2.x-p1.x);
	m1.y = p1.y+(7.0/24.0)*(p2.y-p1.y);
	m2.x = p1.x+(17.0/24.0)*(p2.x-p1.x);
	m2.y = p1.y+(17.0/24.0)*(p2.y-p1.y);
	m3.x = p3.x+(7.0/24.0)*(p4.x-p3.x);
	m3.y = p3.y+(7.0/24.0)*(p4.y-p3.y);
	m4.x = p3.x+(17.0/24.0)*(p4.x-p3.x);
	m4.y = p3.y+(17.0/24.0)*(p4.y-p3.y);

	pos1.x = float(m1.x+(7.0/24.0)*(m4.x-m1.x));
	pos1.y = float(m1.y+(7.0/24.0)*(m4.y-m1.y));
	pos2.x = float(m2.x+(7.0/24.0)*(m3.x-m2.x));
	pos2.y = float(m2.y+(7.0/24.0)*(m3.y-m2.y));
	pos3.x = float(m2.x+(17.0/24.0)*(m3.x-m2.x));
	pos3.y = float(m2.y+(17.0/24.0)*(m3.y-m2.y));
	pos4.x = float(m1.x+(17.0/24.0)*(m4.x-m1.x));
	pos4.y = float(m1.y+(17.0/24.0)*(m4.y-m1.y));

	for (int c_i = 0 ; c_i < circles.size() ; c_i++)
	{
		if (draw_circle_origos)
		{
			cv::circle(drawing, cv::Point2i(circles[c_i].center), 5, cv::Scalar(0,255,0), 1);
			img_proc::show_image("Calculated origo",drawing);
		}


		cv::Point2f new_p1, new_p2, new_p3, new_p4;

		int origo_offset = 3;
		if ((pos1.x-origo_offset < circles[c_i].center.x) && (pos1.x+origo_offset > circles[c_i].center.x) && (pos1.y-origo_offset < circles[c_i].center.y) && (pos1.y+origo_offset > circles[c_i].center.y))
		{
			orient_circ.center = circles[c_i].center;
			orient_circ.circle = circles[c_i].circle;
			return true;
		}
		else if ((pos2.x-origo_offset < circles[c_i].center.x) && (pos2.x+origo_offset > circles[c_i].center.x) && (pos2.y-origo_offset < circles[c_i].center.y) && (pos2.y+origo_offset > circles[c_i].center.y))
		{
			new_p1 = p2;
			new_p2 = p3;
			new_p3 = p4;
			new_p4 = p1;
			p1 = new_p1;
			p2 = new_p2;
			p3 = new_p3;
			p4 = new_p4;
			orient_circ.center = circles[c_i].center;
			orient_circ.circle = circles[c_i].circle;
			return true;
		}
		else if ((pos3.x-origo_offset < circles[c_i].center.x) && (pos3.x+origo_offset > circles[c_i].center.x) && (pos3.y-origo_offset < circles[c_i].center.y) && (pos3.y+origo_offset > circles[c_i].center.y))
		{
			new_p1 = p3;
			new_p2 = p4;
			new_p3 = p1;
			new_p4 = p2;
			p1 = new_p1;
			p2 = new_p2;
			p3 = new_p3;
			p4 = new_p4;
			orient_circ.center = circles[c_i].center;
			orient_circ.circle = circles[c_i].circle;
			return true;
		}
		else if ((pos4.x-origo_offset < circles[c_i].center.x) && (pos4.x+origo_offset > circles[c_i].center.x) && (pos4.y-origo_offset < circles[c_i].center.y) && (pos4.y+origo_offset > circles[c_i].center.y))
		{
			new_p1 = p4;
			new_p2 = p1;
			new_p3 = p2;
			new_p4 = p3;
			p1 = new_p1;
			p2 = new_p2;
			p3 = new_p3;
			p4 = new_p4;
			orient_circ.center = circles[c_i].center;
			orient_circ.circle = circles[c_i].circle;
			return true;
		}
	}
	return false;
}

void markers::identifier::calculate_bit_corners(std::vector<std::vector<cv::Point2f>>& bit_corners, std::vector<cv::Point2f> marker_corners)
{
	std::vector<cv::Point2f> first_bit;
	std::vector<cv::Point2f> second_bit;
	std::vector<cv::Point2f> third_bit;	

	cv::Point2f p1 = marker_corners[0];
	cv::Point2f p2 = marker_corners[1];
	cv::Point2f p3 = marker_corners[2];
	cv::Point2f p4 = marker_corners[3];

	cv::Point2f c1 = cv::Point2f(p1.x+0.5*(p4.x-p1.x),p1.y+0.5*(p4.y-p1.y));
	cv::Point2f c2 = cv::Point2f(p3.x+0.5*(p2.x-p3.x),p3.y+0.5*(p2.y-p3.y));

	cv::Point2f f1 = cv::Point2f(p1.x+(1.0/12.0)*(p4.x-p1.x),p1.y+(1.0/12.0)*(p4.y-p1.y));
	cv::Point2f f2 = cv::Point2f(p1.x+(11.0/12.0)*(p4.x-p1.x),p1.y+(11.0/12.0)*(p4.y-p1.y));
	cv::Point2f f3 = cv::Point2f(p3.x+(1.0/12.0)*(p2.x-p3.x),p3.y+(1.0/12.0)*(p2.y-p3.y));
	cv::Point2f f4 = cv::Point2f(p3.x+(11.0/12.0)*(p2.x-p3.x),p3.y+(11.0/12.0)*(p2.y-p3.y));
	
	cv::Point2f b11 = cv::Point2f(int(f1.x+0.5*(f4.x-f1.x)),int(f1.y+0.5*(f4.y-f1.y)));
	cv::Point2f b12 = cv::Point2f(int(c1.x+0.5*(c2.x-c1.x)),int(c1.y+0.5*(c2.y-c1.y)));
	cv::Point2f b13 = cv::Point2f(int(c1.x+(11.0/12.0)*(c2.x-c1.x)),int(c1.y+(11.0/12.0)*(c2.y-c1.y)));
	cv::Point2f b14 = cv::Point2f(int(f1.x+(11.0/12.0)*(f4.x-f1.x)),int(f1.y+(11.0/12.0)*(f4.y-f1.y)));

	first_bit.push_back(b11);
	first_bit.push_back(b12);
	first_bit.push_back(b13);
	first_bit.push_back(b14);

	cv::Point2f b21 = b12;
	cv::Point2f b22 = cv::Point2f(int(f2.x+0.5*(f3.x-f2.x)),int(f2.y+0.5*(f3.y-f2.y)));
	cv::Point2f b23 = cv::Point2f(int(f2.x+(11.0/12.0)*(f3.x-f2.x)),int(f2.y+(11.0/12.0)*(f3.y-f2.y)));
	cv::Point2f b24 = b13;

	second_bit.push_back(b21);
	second_bit.push_back(b22);
	second_bit.push_back(b23);
	second_bit.push_back(b24);

	cv::Point2f b31 = cv::Point2f(int(c1.x+(1.0/12.0)*(c2.x-c1.x)),int(c1.y+(1.0/12.0)*(c2.y-c1.y)));
	cv::Point2f b32 = cv::Point2f(int(f2.x+(1.0/12.0)*(f3.x-f2.x)),int(f2.y+(1.0/12.0)*(f3.y-f2.y)));
	cv::Point2f b33 = b22;
	cv::Point2f b34 = b21;

	third_bit.push_back(b31);
	third_bit.push_back(b32);
	third_bit.push_back(b33);
	third_bit.push_back(b34);

	bit_corners.push_back(first_bit);
	bit_corners.push_back(second_bit);
	bit_corners.push_back(third_bit);
}

void markers::identifier::draw_contour(std::vector<cv::Point> contour)
{
	for (int i = 0; i < contour.size() ; i++)
	{
		if (i == contour.size()-1)
		{
			cv::line(drawing, contour[i], contour[0], cv::Scalar(255,0,255), 4);
		}
		else
		{
			cv::line(drawing, contour[i], contour[i+1], cv::Scalar(255,0,255), 4);
		}	
	}
}

bool markers::identifier::check_orientation_circle_contents(orientation_circle_t circle)
{
	bool draw_square = false;

	cv::Mat image = THRESH.threshold_image.clone();
	cv::Rect rect;

	rect = boundingRect(circle.circle);

	float contour_area = cv::contourArea(circle.circle);
	float rectangle_area = rect.height*rect.width;

	//50-70?
	float num_pixels = 0.0;
	double pixel_intensity = 0.0;
	for (int x = rect.x; x < rect.x+rect.width; x++)
	{
		for (int y = rect.y; y < rect.y+rect.height; y++)
		{
			num_pixels++;
			pixel_intensity += image.at<uchar>(cv::Point2i(x,y));
		}
	}

	if (pixel_intensity/num_pixels < 80)
	{
		return true;
	}
	else
	{
		return false;
	}
}

*/




//Marker decoder
markers::decoder::decoder(threshold_t thresh)
{
	THRESH = thresh;
	drawing = thresh.original_image.clone(); 
	threshold_img = THRESH.threshold_image.clone();
}

threshold_t markers::decoder::decode_marker(void)
{
	possible_marker_t current_marker;
	int bit1, bit2, bit3;

	std::vector<int> delete_list;

	for (int pm_i = 0 ; pm_i < THRESH.possible_markers.size() ; pm_i++)
	{
		bit1 = -1;
		bit2 = -1;
		bit3 = -1;

		current_marker = THRESH.possible_markers[pm_i];

		int shortest_edge = fmax(THRESH.original_image.size().width,THRESH.original_image.size().height);

		cv::Point2f pt1, pt2;

		for (int j = 0 ; j < current_marker.corners.size() ; j++)
		{
			if (j == 0)
			{
				pt1 = current_marker.corners[current_marker.corners.size()-1];
				pt2 = current_marker.corners[j];
			}
			else
			{
				pt1 = current_marker.corners[j-1];
				pt2 = current_marker.corners[j];
			}
			
			int edge_size = std::sqrt(pow(pt2.x-pt1.x,2)+pow(pt2.y-pt1.y,2));
			if (edge_size < shortest_edge)
			{
				shortest_edge = edge_size;
			}
		}

		int smallest_bit_size = int(((1.0/2.0)-(1.0/12.0))*shortest_edge);
		
		std::vector<cv::Point2f> First_bit_corners = current_marker.bit_corners[0];
		std::vector<cv::Point2f> Second_bit_corners = current_marker.bit_corners[1];
		std::vector<cv::Point2f> Third_bit_corners = current_marker.bit_corners[2];

		int loop_counter = 0;
		while (bit1 == -1)
		{
			bit1 = decode_bit(First_bit_corners,smallest_bit_size);
			loop_counter+=1;
			if (loop_counter == 5)
			{
				break;
			}
		}
		loop_counter = 0;
		while (bit2 == -1)
		{
			bit2 = decode_bit(Second_bit_corners,smallest_bit_size);
			loop_counter+=1;
			if (loop_counter == 5)
			{
				break;
			}
		}
		loop_counter = 0;
		while (bit3 == -1)
		{
			bit3 = decode_bit(Third_bit_corners,smallest_bit_size);
			loop_counter+=1;
			if (loop_counter == 5)
			{
				break;
			}
		}

		if ((bit1 == -1) || (bit2 == -1) || (bit3 == -1))
		{
			delete_list.push_back(pm_i);
		}
		else
		{	
			THRESH.possible_markers[pm_i].code = (100*bit1)+(10*bit2)+(bit3);
		}
	}

	for (int dl_i = delete_list.size()-1 ; dl_i >= 0 ; dl_i--)
	{
		THRESH.possible_markers.erase(THRESH.possible_markers.begin()+delete_list[dl_i]);
	}
	return THRESH;
}

int markers::decoder::decode_bit(std::vector<cv::Point2f> bit_corners, int smallest_bit_size)
{
	cv::Point2f center_point;
	int x_sum = 0;
	int y_sum = 0;
	
	for (int i = 0 ; i < bit_corners.size() ; i++)
	{
		x_sum += bit_corners[i].x;
		y_sum += bit_corners[i].y;
	}
	center_point = cv::Point2f(int(x_sum/int(bit_corners.size())),int(y_sum/int(bit_corners.size())));

	cv::Rect rect = cv::boundingRect(bit_corners);

	double pixel_intensity = 0.0;
	int pixel_counter = 0;
	for (int x = rect.x; x < rect.x+rect.width; x++)
	{
		for (int y = rect.y; y < rect.y+rect.height; y++)
		{
			pixel_intensity += threshold_img.at<uchar>(cv::Point2i(x,y));
			++pixel_counter;
		}
	}
	double mean_pixel_value = pixel_intensity/pixel_counter;

	
	if (mean_pixel_value < 70.0)
	{
		return 1;
	}
	else if (mean_pixel_value > 170.0) 
	{
		return 0;
	}
	else
	{
		return -1;
	}
}


//Extra functions
double angle(cv::Point pt1, cv::Point pt2, cv::Point pt0)
{
	double dx1 = pt1.x - pt0.x;
	double dy1 = pt1.y - pt0.y;
	double dx2 = pt2.x - pt0.x;
	double dy2 = pt2.y - pt0.y;

	return (dx1*dx2 + dy1*dy2)/sqrt((dx1*dx1 + dy1*dy1)*(dx2*dx2 + dy2*dy2) + 1e-10);
}

std::vector<marker_t> evaluate_results(std::vector<threshold_t> thresh)
{	
	std::vector<marker_t> return_markers;
	for (int i = 0 ; i < 2 ; ++i)
	{
		for (int j = 0 ; j < 2 ; ++j)
		{
			for (int k = 0 ; k < 2 ; ++k)
			{
				marker_t marker;
				std::vector<cv::Point2i> corner1, corner2, corner3, corner4;
				cv::Rect rect1, rect2, rect3, rect4;
				int num_detections = 0; 

				for (int t=0; t<thresh.size(); ++t)
				{
					for (int m=0; m<thresh[t].possible_markers.size(); ++m)
					{
						if (thresh[t].possible_markers[m].code == ((100*i)+(10*j)+k))
						{
							++num_detections;
							corner1.push_back(thresh[t].possible_markers[m].corners[0]);
							corner2.push_back(thresh[t].possible_markers[m].corners[1]);
							corner3.push_back(thresh[t].possible_markers[m].corners[2]);
							corner4.push_back(thresh[t].possible_markers[m].corners[3]);
						}
					}
				}
				rect1 = cv::boundingRect(corner1);
				rect2 = cv::boundingRect(corner2);
				rect3 = cv::boundingRect(corner3);
				rect4 = cv::boundingRect(corner4);
				if ((rect1.width > 4) || (rect1.height > 4) || (rect2.width > 4) || (rect2.height > 4) || (rect3.width > 4) || (rect3.height > 4) || (rect4.width > 4) || (rect4.height > 4))
				{
					std::cout << "Need to get rid of extreme points." << std::endl;
				}
				if ((corner1.size()>0) && (corner2.size()>0) && (corner3.size()>0) && (corner4.size()>0))
				{
					marker.corners.push_back(cv::Point2i(rect1.x+(.5*rect1.width), rect1.y+(.5*rect1.height)));
					marker.corners.push_back(cv::Point2i(rect2.x+(.5*rect2.width), rect2.y+(.5*rect2.height)));
					marker.corners.push_back(cv::Point2i(rect3.x+(.5*rect3.width), rect3.y+(.5*rect3.height)));
					marker.corners.push_back(cv::Point2i(rect4.x+(.5*rect4.width), rect4.y+(.5*rect4.height)));
					marker.code = ((100*i)+(10*j)+k);
					return_markers.push_back(marker);
				}
			}
		}
	}
	return return_markers;
}
