
#include<stdio.h>
#include"opencv2/opencv.hpp"

#include"../headers/pose_estimation.hpp"
#include"../headers/marker_detection.hpp"
#include"../headers/position_heading_calculation.hpp"
#include"../headers/printing.hpp"


bool estimate_pose(std::vector<cv::Point2f> img_pts, std::vector<cv::Point3f> real_pts, cv::Mat& transformationScene2Camera)
{
	bool ret = false;

	std::string datafile = "configuration_files/cameraData.xml";
	cv::FileStorage fs(datafile, cv::FileStorage::READ);

	cv::Mat distCoeffs;
	cv::Mat cam_mtx(3,3,CV_64F);

	fs["camera_matrix"] >> cam_mtx;
	fs["distortion_coefficients"] >> distCoeffs;

	cv::Mat rvec(3,1,CV_64F);
	cv::Mat tvec(3,1,CV_64F);
/*
	std::cout << "Real points size: " << real_pts.size() <<  std::endl;
	for (int i = 0; i < real_pts.size(); i++)
	{
		std::cout << real_pts[i] << std::endl;
	}
	*/

	ret = cv::solvePnP(real_pts, img_pts, cam_mtx, distCoeffs, rvec, tvec, false, cv::SOLVEPNP_ITERATIVE);
	
	if (ret)
	{
		cv::Mat rmat;
		cv::Rodrigues(rvec, rmat);
		/*
		cv::Mat tmat = cv::Mat(tvec);
		pf::printTransformationMatrix(tvec);
		rmat.convertTo(rmat, CV_32F);
		tmat.convertTo(tmat, CV_32F);
		pf::printTransformationMatrix(tmat);
		

		std::cout << "rmat type is: " << type2str(rmat.type()) << std::endl;
		std::cout << "tmat type is: " << type2str(tmat.type()) << std::endl;
		*/

		for (int r = 0; r<rmat.rows; ++r)
		{
			for (int c = 0; c<rmat.cols; ++c)
			{
				transformationScene2Camera.at<double>(r,c) = rmat.at<double>(r,c);
			}
			transformationScene2Camera.at<double>(r,3) = tvec.at<double>(r,0);
		}
		for (int i = 0; i<3; ++i)
		{
			transformationScene2Camera.at<double>(3,i) = 0;
		}
		transformationScene2Camera.at<double>(3,3) = 1;
	}

	return ret;
}

bool get_real_pts(std::string quay_name, std::vector<marker_t>& detected_markers, std::vector<cv::Point2f>& img_pts, std::vector<cv::Point3f>& real_pts, std::vector<int>& detected_codes)
{
	/*
	Based on the quay and the detected points of the markers, this function will save both the detected points of the markers in the image as well as the corresponding points in the real world. Based on the correct configuration file for the specific quay. 
	*/
	std::vector<marker_real_t> quay_markers;
	bool load_success = false;
	bool already_done = false;
	marker_t current_detected_marker;
	marker_real_t current_real_marker;
	//std::cout << "Starting to load quay information" << std::endl;
	load_success = load_quay_information(quay_name, quay_markers);
	//std::cout << "Done loading the quay information" << std::endl;
	if (load_success && (detected_markers.size() > 0))
	{
		for (int index_1 = 0 ; index_1 < detected_markers.size() ; index_1++)
		{
			current_detected_marker = detected_markers[index_1];

			for (int code_index = 0 ; code_index < detected_codes.size() ; code_index++)
			{
				if (detected_codes[code_index] == current_detected_marker.code)
				{
					already_done = true;
					for (int index_2 = 0; index_2 < 4; ++index_2)
					{
						img_pts[(code_index*4)+index_2] = current_detected_marker.corners[index_2];
					}
					break;
				}
			}
			if (!already_done)
			{
				for (int index_2 = 0 ; index_2 < quay_markers.size() ; index_2++)
				{
					current_real_marker = quay_markers[index_2];
					if (current_real_marker.code == current_detected_marker.code)
					{
						detected_codes.push_back(current_detected_marker.code);

						for (int index_3 = 0 ; index_3 < current_real_marker.pts.size() ; index_3++)
						{
							img_pts.push_back(current_detected_marker.corners[index_3]);
							real_pts.push_back(current_real_marker.pts[index_3]);
						}
					}
				}
			}
		}
	}
	return true;
}

bool load_quay_information(std::string quay_name, std::vector<marker_real_t>& real_pts)
{
	marker_real_t current_marker;

	std::string datafile = "configuration_files/quay_config_files/"+quay_name+".xml";
	cv::FileStorage quay(datafile, cv::FileStorage::READ);

	for (int i = 0; i < 2; ++i)
	{
		for (int j = 0; j < 2; ++j)
		{
			for (int k = 0; k < 2; ++k)
			{
				cv::Mat current_point; 
				current_marker.code = 100*i+10*j+k;

				for (int point = 1; point < 5; ++point)
				{
					std::string current_string = "marker_"+std::to_string(i)+std::to_string(j)+std::to_string(k)+"_pt"+std::to_string(point);
					quay[current_string] >> current_point;
					current_marker.pts.push_back(cv::Point3f(current_point.at<double>(0,0),current_point.at<double>(0,1),current_point.at<double>(0,2)));
				}
				real_pts.push_back(current_marker);
				current_marker = {};
			}
		}
	}
	if (real_pts.size() > 0)
	{
		return true;
	}
	else
	{
		return false;
	}
}

std::string type2str(int type) {
  std::string r;

  uchar depth = type & CV_MAT_DEPTH_MASK;
  uchar chans = 1 + (type >> CV_CN_SHIFT);

  switch ( depth ) {
    case CV_8U:  r = "8U"; break;
    case CV_8S:  r = "8S"; break;
    case CV_16U: r = "16U"; break;
    case CV_16S: r = "16S"; break;
    case CV_32S: r = "32S"; break;
    case CV_32F: r = "32F"; break;
    case CV_64F: r = "64F"; break;
    default:     r = "User"; break;
  }

  r += "C";
  r += (chans+'0');

  return r;
}