#include<stdio.h>
#include<cmath>
#include<math.h>
#include"opencv2/opencv.hpp"

#include"../headers/position_heading_calculation.hpp"
#include"../headers/printing.hpp"
#include"../headers/pose_estimation.hpp"

position_t calculate_position_heading(cv::Mat Trans_quay_camera)
{
	position_t ship_pose;
	
	//Load ferry information
	std::string ferry_data_file = "configuration_files/ferry_config_files/NTNU-tests.xml";
	cv::FileStorage ferry_storage(ferry_data_file,cv::FileStorage::READ);

	cv::Mat transformation_camera_ship;
	ferry_storage["transformation_camera_ship"] >> transformation_camera_ship;


	ferry_storage.release();

	//Load quay information
	std::string quay_data_file = "configuration_files/quay_config_files/single_straight_marker_tests.xml";
	cv::FileStorage quay_storage(quay_data_file,cv::FileStorage::READ);

	cv::Mat quay_position;
	quay_storage["position"] >> quay_position;

	quay_storage.release();

	//(Latitude, Longitude, altitude) = (phi, lambda, h)
	cv::Vec4f Quay_ECEF_position(0.,0.,0.,1.0);
	geotoECEF(quay_position, Quay_ECEF_position);

	pf::print_ECEF_position(Quay_ECEF_position, "quay");

	cv::Mat transformation_ECEF_Q = cv::Mat::eye(4,4,CV_64F);
	getTransECEFtoQ(Quay_ECEF_position, transformation_ECEF_Q);
	
	calculate_position(ship_pose, transformation_ECEF_Q, Trans_quay_camera, transformation_camera_ship);

	calculate_heading(ship_pose);

	return ship_pose;
}


void calculate_position(position_t& ship_pose,cv::Mat Trans_ECEF_quay, cv::Mat Trans_quay_camera, cv::Mat Trans_camera_ship)
{
	cv::Vec4f point_cam, point_quay, point_ECEF;

	point_cam = change_point_coordinate_frame(Trans_camera_ship, cv::Vec4f(0.0,0.0,0.0,1.0));
	point_quay = change_point_coordinate_frame(Trans_quay_camera, point_cam);
	point_ECEF = change_point_coordinate_frame(Trans_ECEF_quay, point_quay);

	pf::print_ECEF_position(point_ECEF, "ship");

	ECEFtoGeo(point_ECEF, ship_pose);	
}

void calculate_heading(position_t& ship_pose)
{
	std::cout << "Calculate Heading" << std::endl;
	ship_pose.Heading = 100.0;
}

cv::Vec4f change_point_coordinate_frame(cv::Mat transformation, cv::Vec4f point)
{
	
	cv::Vec4f new_point(0,0,0,0);
	for (int i = 0; i < transformation.size().height; ++i)
	{
		double value = 0; 
		for (int j = 0; j < transformation.size().width; ++j)
		{
			value += transformation.at<double>(i,j)*point[j];
		}
		new_point[i] = value;
	}
	return new_point;
}

void geotoECEF(cv::Mat Quay_geodetic_position, cv::Vec4f& Quay_ECEF_position)
{
	double phi = Quay_geodetic_position.at<double>(0,0)*(CV_PI/180);
	double lambda = Quay_geodetic_position.at<double>(0,1)*(CV_PI/180);
	double altitude = Quay_geodetic_position.at<double>(0,2);

	double a = 6378137;
	double b = 6356752;

	double e_squared = 1-(std::pow(b,2)/std::pow(a,2));

	double N = a/(sqrt(1-(e_squared*std::pow(sin(phi),2))));

	Quay_ECEF_position[0] = (N+altitude)*cos(phi)*cos(lambda);
	Quay_ECEF_position[1] = (N+altitude)*cos(phi)*sin(lambda);
	Quay_ECEF_position[2] = (((std::pow(b,2)/std::pow(a,2))*N)+altitude)*sin(phi);
	Quay_ECEF_position[3] = 1.;
}

void getTransECEFtoQ(cv::Vec4f Quay_ECEF_position, cv::Mat& trans)
{
	cv::Mat inverse_trans = cv::Mat::eye(4,4,CV_64F);

	double alpha = -atan2(Quay_ECEF_position[0],Quay_ECEF_position[1]);
	double beta = -atan2(Quay_ECEF_position[2],sqrt(std::pow(Quay_ECEF_position[0],2)+std::pow(Quay_ECEF_position[1],2)));

	cv::Mat Rot_x = (cv::Mat_<double>(3,3) << 1, 0, 0, 0, cos(beta),-sin(beta), 0, sin(beta), cos(beta));
	cv::Mat Rot_z = (cv::Mat_<double>(3,3) << cos(alpha),-sin(alpha), 0, sin(alpha), cos(alpha),0,0,0,1);

	cv::Mat R = Rot_x*Rot_z;

	for (int r = 0; r < inverse_trans.size().height-1; ++r)
	{
		for (int c = 0; c < R.size().width; ++c)
		{
			inverse_trans.at<double>(r,c) = R.at<double>(r,c);
		}
		inverse_trans.at<double>(r,3) = Quay_ECEF_position[r];
	}

	trans = inverse_trans.inv();
}

void ECEFtoGeo(cv::Vec4f ECEF_point, position_t& ship_pose)
{
	double a = 6378137;
	double b = 6356752;
	double e1_squared = 1-(std::pow(b,2)/std::pow(a,2));

	double x = ECEF_point[0];
	double y = ECEF_point[1];
	double z = ECEF_point[2];
	
	double r = sqrt(std::pow(x,2)+std::pow(y,2));
	double e2_squared = (std::pow(a,2)-std::pow(b,2))/std::pow(b,2);
	double f = 54*std::pow(b,2)*std::pow(z,2);
	double g = std::pow(r,2)+((1-e1_squared)*std::pow(z,2))-(e1_squared*(std::pow(1,2)-std::pow(b,2)));
	double c = (std::pow(e1_squared,2)*f*std::pow(r,2))/(std::pow(g,3));
	double s = std::cbrt(1+c+sqrt(std::pow(c,2)+(2*c)));
	double p = f/(3*std::pow(s+(1/s)+1,2)*std::pow(g,2));
	double q = sqrt(1+(2*std::pow(e1_squared,2)*p));
	double r0 = (-(p*e1_squared*r)/(1+q))+sqrt((.5*std::pow(a,2)*(1+(1/q)))-((p*(1-e1_squared)*std::pow(z,2))/(q*(1+q)))-(.5*p*std::pow(r,2)));
	double u = sqrt(std::pow(r-(e1_squared*r0),2)+std::pow(z,2));
	double v = sqrt(std::pow(r-(e1_squared*r0),2)+((1-e1_squared)*std::pow(z,2)));
	double z0 = ((std::pow(b,2)*z)/(a*v));
	double h = u*(1-(std::pow(b,2)/(a*v)));
	double phi = atan((z+(e2_squared*z0))/r);
	double lambda = atan2(y,x);

	ship_pose.Latitude = phi*(180/CV_PI);
	ship_pose.Longitude = lambda*(180/CV_PI);
	ship_pose.Altitude = h;
}

void calculatePose(pose_t& camera_pose, cv::Mat transformationScene2Camera)
{
	cv::Mat_<double> transformationCamera2Scene = transformationScene2Camera.inv();

	cv::Mat_<double> POc = cv::Mat(cv::Vec4f(0,0,0,1), CV_64F);
	cv::Mat_<double> PZc = cv::Mat(cv::Vec4f(0,0,1,1), CV_64F);
	cv::Mat POs = transformationCamera2Scene*POc;
	cv::Mat PZs = transformationCamera2Scene*PZc;

	double over, under;
	if ((POs.at<double>(0,0)>=0) && (PZs.at<double>(0,0)>=0))
	{
		over = PZs.at<double>(0,0)-POs.at<double>(0,0);
	}
	else if ((POs.at<double>(0,0)<0) && (PZs.at<double>(0,0)<0))
	{
		over = PZs.at<double>(0,0)-POs.at<double>(0,0);
	}
	else if ((POs.at<double>(0,0)>=0) && (PZs.at<double>(0,0)<0))
	{
		over = -(fabs(POs.at<double>(0,0))+fabs(PZs.at<double>(0,0)));
	}
	else if ((POs.at<double>(0,0)<0) && (PZs.at<double>(0,0)>=0))
	{
		over = (fabs(POs.at<double>(0,0))+fabs(PZs.at<double>(0,0)));
	}

	under = POs.at<double>(0,2)-PZs.at<double>(0,2); 

	//std::cout << "Over is: " << over << std::endl;
	//std::cout << "Under is: " << under << std::endl;

	camera_pose.psi = atan(over/under);
	camera_pose.x = POs.at<double>(0,0);
	camera_pose.z = POs.at<double>(0,2);

}





