#include<stdio.h>
#include<iostream>

#include"opencv2/opencv.hpp"

#include"../headers/marker_detection.hpp"
#include"../headers/printing.hpp"
#include"../headers/position_heading_calculation.hpp"
#include"../headers/pose_estimation.hpp"


void pf::print_marker(marker_t marker)
{
	std::cout << "\nMarker " << marker.code << " located." << std::endl;
	std::cout << "The corners are located in the following positions in the image: " << std::endl;
	for (int i = 0 ; i < marker.corners.size() ; i++)
	{
		std::cout << "\t Point " << i+1 << ": " << marker.corners[i] << std::endl;
	}
}


void pf::print_position_result(position_t position)
{
	std::cout << "The GNSS coordinates of the ship is currently: " << std::endl;
	std::cout << "Lat.: " << position.Latitude << std::endl;
	std::cout << "Long.: " << position.Longitude << std::endl;
	std::cout << "Altitude: " << position.Altitude << std::endl;
	std::cout << "Heading: " << position.Heading << std::endl;
}

void pf::print_ECEF_position(cv::Vec4f position, std::string name)
{
	std::cout << "The " << name << " has the following ECEF position: " << std::endl;
	std::cout << "X: " << position[0] << std::endl;
	std::cout << "Y: " << position[1] << std::endl;
	std::cout << "Z: " << position[2] << std::endl;
	std::cout << "Extra coordinate: " << position[3] << std::endl;
}

cv::Mat pf::print_marker_detected(std::vector<marker_t> markers,cv::Mat image)
{
	bool should_save = false;

	cv::Mat drawing = image.clone();

	for (int marker_i = 0; marker_i < markers.size(); ++marker_i)
	{
		marker_t mrk = markers[marker_i];

		for (int point_i = 0; point_i < mrk.corners.size(); ++ point_i)
		{
			if (point_i < mrk.corners.size()-1)
			{
				cv::line(drawing, mrk.corners[point_i], mrk.corners[point_i+1], cv::Scalar(0,0,255),3);
			}
			else
			{
				cv::line(drawing, mrk.corners[point_i], mrk.corners[0], cv::Scalar(0,0,255),3);
				cv::circle(drawing, mrk.corners[0], 10, cv::Scalar(255,0,0),-1);
			}
			cv::circle(drawing, mrk.corners[point_i], 10, cv::Scalar(255,0,0),-1);
		}
		cv::putText(drawing,std::to_string(mrk.code),cv::Point2i(mrk.corners[1].x+20,mrk.corners[1].y-20), cv::FONT_HERSHEY_SIMPLEX, 4, cv::Scalar(255,255,255), 2, cv::LINE_AA);
	}

	return drawing;

}


void pf::printTransformationMatrix(cv::Mat matrix)
{
	std::cout << "Transformation is: " << std::endl;
	for (int i = 0; i < matrix.rows; ++i)
	{
		for (int j = 0; j < matrix.cols; ++j)
		{
			std::cout << matrix.at<double>(i,j) << "	";
		}
		std::cout << std::endl;
	}
}

void pf::printPose(pose_t pose)
{
	std::cout << "The pose of the camera is: " << std::endl;
	std::cout << "	x: " << pose.x << " meters" << std::endl;
	std::cout << "	z: " << pose.z << " meters" << std::endl;
	std::cout << "	psi: " << pose.psi*(180/CV_PI) << " degrees" << std::endl;
}

void pf::printPoint(cv::Mat point)
{
	std::cout << "Point is: " << std::endl;
	std::cout << "	x: " << point.at<double>(0,0) << std::endl;
	std::cout << "	y: " << point.at<double>(0,1) << std::endl;
	std::cout << "	z: " << point.at<double>(0,2) << std::endl;
}
